@@ -21,32 +21,101 @@ import (
 	"testing"
 	"time"
 
-	"go.etcd.io/etcd/pkg/flags"
-	"go.etcd.io/etcd/pkg/testutil"
-	"go.etcd.io/etcd/version"
+	"github.com/stretchr/testify/assert"
+	"go.etcd.io/etcd/api/v3/version"
+	"go.etcd.io/etcd/client/pkg/v3/fileutil"
+	"go.etcd.io/etcd/client/pkg/v3/testutil"
+	"go.etcd.io/etcd/pkg/v3/flags"
+	"go.etcd.io/etcd/tests/v3/framework/e2e"
 )
 
 func TestCtlV3Version(t *testing.T) { testCtl(t, versionTest) }
 
+func TestClusterVersion(t *testing.T) {
+	e2e.BeforeTest(t)
+
+	tests := []struct {
+		name         string
+		rollingStart bool
+	}{
+		{
+			name:         "When start servers at the same time",
+			rollingStart: false,
+		},
+		{
+			name:         "When start servers one by one",
+			rollingStart: true,
+		},
+	}
+
+	for _, tt := range tests {
+		t.Run(tt.name, func(t *testing.T) {
+			binary := e2e.BinDir + "/etcd"
+			if !fileutil.Exist(binary) {
+				t.Skipf("%q does not exist", binary)
+			}
+			e2e.BeforeTest(t)
+			cfg := e2e.NewConfigNoTLS()
+			cfg.ExecPath = binary
+			cfg.SnapshotCount = 3
+			cfg.BaseScheme = "unix" // to avoid port conflict
+			cfg.RollingStart = tt.rollingStart
+
+			epc, err := e2e.NewEtcdProcessCluster(t, cfg)
+			if err != nil {
+				t.Fatalf("could not start etcd process cluster (%v)", err)
+			}
+			defer func() {
+				if errC := epc.Close(); errC != nil {
+					t.Fatalf("error closing etcd processes (%v)", errC)
+				}
+			}()
+
+			ctx := ctlCtx{
+				t:   t,
+				cfg: *cfg,
+				epc: epc,
+			}
+			cv := version.Cluster(version.Version)
+			clusterVersionTest(ctx, `"etcdcluster":"`+cv)
+		})
+	}
+}
+
 func versionTest(cx ctlCtx) {
 	if err := ctlV3Version(cx); err != nil {
 		cx.t.Fatalf("versionTest ctlV3Version error (%v)", err)
 	}
 }
 
-func ctlV3Version(cx ctlCtx) error {
-	cmdArgs := append(cx.PrefixArgs(), "version")
-	return spawnWithExpect(cmdArgs, version.Version)
+func clusterVersionTest(cx ctlCtx, expected string) {
+	var err error
+	for i := 0; i < 35; i++ {
+		if err = e2e.CURLGet(cx.epc, e2e.CURLReq{Endpoint: "/version", Expected: expected}); err != nil {
+			cx.t.Logf("#%d: v3 is not ready yet (%v)", i, err)
+			time.Sleep(200 * time.Millisecond)
+			continue
+		}
+		break
+	}
+	if err != nil {
+		cx.t.Fatalf("failed cluster version test expected %v got (%v)", expected, err)
+	}
 }
 
-// TestCtlV3DialWithHTTPScheme ensures that client handles endpoints with HTTPS scheme.
+func ctlV3Version(cx ctlCtx) error {
+	cmdArgs := append(cx.PrefixArgs(), "version")
+	return e2e.SpawnWithExpectWithEnv(cmdArgs, cx.envMap, version.Version)
+}
+
+// TestCtlV3DialWithHTTPScheme ensures that client handles Endpoints with HTTPS scheme.
 func TestCtlV3DialWithHTTPScheme(t *testing.T) {
-	testCtl(t, dialWithSchemeTest, withCfg(configClientTLS))
+	testCtl(t, dialWithSchemeTest, withCfg(*e2e.NewConfigClientTLS()))
 }
 
 func dialWithSchemeTest(cx ctlCtx) {
 	cmdArgs := append(cx.prefixArgs(cx.epc.EndpointsV3()), "put", "foo", "bar")
-	if err := spawnWithExpect(cmdArgs, "OK"); err != nil {
+	if err := e2e.SpawnWithExpectWithEnv(cmdArgs, cx.envMap, "OK"); err != nil {
 		cx.t.Fatal(err)
 	}
 }
@@ -54,14 +123,14 @@ func dialWithSchemeTest(cx ctlCtx) {
 type ctlCtx struct {
 	t                 *testing.T
 	apiPrefix         string
-	cfg               etcdProcessClusterConfig
+	cfg               e2e.EtcdProcessClusterConfig
 	quotaBackendBytes int64
 	corruptFunc       func(string) error
 	noStrictReconfig  bool
 
-	epc *etcdProcessCluster
+	epc *e2e.EtcdProcessCluster
 
-	envMap map[string]struct{}
+	envMap map[string]string
 
 	dialTimeout time.Duration
 	testTimeout time.Duration
@@ -76,6 +145,12 @@ type ctlCtx struct {
 
 	// for compaction
 	compactPhysical bool
+
+	// to run etcdutl instead of etcdctl for suitable commands.
+	etcdutl bool
+
+	// dir that was used during the test
+	dataDir string
 }
 
 type ctlOption func(*ctlCtx)
@@ -87,7 +162,7 @@ func (cx *ctlCtx) applyOpts(opts []ctlOption) {
 	cx.initialCorruptCheck = true
 }
 
-func withCfg(cfg etcdProcessClusterConfig) ctlOption {
+func withCfg(cfg e2e.EtcdProcessClusterConfig) ctlOption {
 	return func(cx *ctlCtx) { cx.cfg = cfg }
 }
 
@@ -132,7 +207,11 @@ func withApiPrefix(p string) ctlOption {
 }
 
 func withFlagByEnv() ctlOption {
-	return func(cx *ctlCtx) { cx.envMap = make(map[string]struct{}) }
+	return func(cx *ctlCtx) { cx.envMap = make(map[string]string) }
+}
+
+func withEtcdutl() ctlOption {
+	return func(cx *ctlCtx) { cx.etcdutl = true }
 }
 
 // This function must be called after the `withCfg`, otherwise its value
@@ -143,52 +222,70 @@ func withMaxConcurrentStreams(streams uint32) ctlOption {
 	}
 }
 
-func withDebug(debug bool) ctlOption {
+// This function must be called after the `withCfg`, otherwise its value
+// may be overwritten by `withCfg`.
+func withSnapshotCount(snapshotCount int) ctlOption {
 	return func(cx *ctlCtx) {
-		cx.cfg.debug = debug
+		cx.cfg.SnapshotCount = snapshotCount
 	}
 }
 
+func withLogLevel(logLevel string) ctlOption {
+	return func(cx *ctlCtx) {
+		cx.cfg.LogLevel = logLevel
+	}
+}
+
+func testCtl(t *testing.T, testFunc func(ctlCtx), opts ...ctlOption) {
+	testCtlWithOffline(t, testFunc, nil, opts...)
+}
+
 func getDefaultCtlCtx(t *testing.T) ctlCtx {
 	return ctlCtx{
 		t:           t,
-		cfg:         configAutoTLS,
+		cfg:         *e2e.NewConfigAutoTLS(),
 		dialTimeout: 7 * time.Second,
 	}
 }
 
-func testCtl(t *testing.T, testFunc func(ctlCtx), opts ...ctlOption) {
-	defer testutil.AfterTest(t)
+func testCtlWithOffline(t *testing.T, testFunc func(ctlCtx), testOfflineFunc func(ctlCtx), opts ...ctlOption) {
+	e2e.BeforeTest(t)
 
 	ret := getDefaultCtlCtx(t)
 	ret.applyOpts(opts)
 
-	mustEtcdctl(t)
 	if !ret.quorum {
-		ret.cfg = *configStandalone(ret.cfg)
+		ret.cfg = *e2e.ConfigStandalone(ret.cfg)
 	}
 	if ret.quotaBackendBytes > 0 {
-		ret.cfg.quotaBackendBytes = ret.quotaBackendBytes
+		ret.cfg.QuotaBackendBytes = ret.quotaBackendBytes
 	}
-	ret.cfg.noStrictReconfig = ret.noStrictReconfig
+	ret.cfg.NoStrictReconfig = ret.noStrictReconfig
 	if ret.initialCorruptCheck {
-		ret.cfg.initialCorruptCheck = ret.initialCorruptCheck
+		ret.cfg.InitialCorruptCheck = ret.initialCorruptCheck
+	}
+	if testOfflineFunc != nil {
+		ret.cfg.KeepDataDir = true
 	}
 
-	epc, err := newEtcdProcessCluster(t, &ret.cfg)
+	epc, err := e2e.NewEtcdProcessCluster(t, &ret.cfg)
 	if err != nil {
 		t.Fatalf("could not start etcd process cluster (%v)", err)
 	}
 	ret.epc = epc
+	ret.dataDir = epc.Procs[0].Config().DataDirPath
 
 	defer func() {
 		if ret.envMap != nil {
 			for k := range ret.envMap {
 				os.Unsetenv(k)
 			}
+			ret.envMap = make(map[string]string)
 		}
-		if errC := ret.epc.Close(); errC != nil {
-			t.Fatalf("error closing etcd processes (%v)", errC)
+		if ret.epc != nil {
+			if errC := ret.epc.Close(); errC != nil {
+				t.Fatalf("error closing etcd processes (%v)", errC)
+			}
 		}
 	}()
 
@@ -196,6 +293,7 @@ func testCtl(t *testing.T, testFunc func(ctlCtx), opts ...ctlOption) {
 	go func() {
 		defer close(donec)
 		testFunc(ret)
+		t.Log("---testFunc logic DONE")
 	}()
 
 	timeout := ret.getTestTimeout()
@@ -205,6 +303,15 @@ func testCtl(t *testing.T, testFunc func(ctlCtx), opts ...ctlOption) {
 		testutil.FatalStack(t, fmt.Sprintf("test timed out after %v", timeout))
 	case <-donec:
 	}
+
+	t.Log("closing test cluster...")
+	assert.NoError(t, epc.Close())
+	epc = nil
+	t.Log("closed test cluster...")
+
+	if testOfflineFunc != nil {
+		testOfflineFunc(ret)
+	}
 }
 
 func (cx *ctlCtx) getTestTimeout() time.Duration {
@@ -222,18 +329,18 @@ func (cx *ctlCtx) prefixArgs(eps []string) []string {
 	fmap := make(map[string]string)
 	fmap["endpoints"] = strings.Join(eps, ",")
 	fmap["dial-timeout"] = cx.dialTimeout.String()
-	if cx.epc.cfg.clientTLS == clientTLS {
-		if cx.epc.cfg.isClientAutoTLS {
+	if cx.epc.Cfg.ClientTLS == e2e.ClientTLS {
+		if cx.epc.Cfg.IsClientAutoTLS {
 			fmap["insecure-transport"] = "false"
 			fmap["insecure-skip-tls-verify"] = "true"
-		} else if cx.epc.cfg.isClientCRL {
-			fmap["cacert"] = caPath
-			fmap["cert"] = revokedCertPath
-			fmap["key"] = revokedPrivateKeyPath
+		} else if cx.epc.Cfg.IsClientCRL {
+			fmap["cacert"] = e2e.CaPath
+			fmap["cert"] = e2e.RevokedCertPath
+			fmap["key"] = e2e.RevokedPrivateKeyPath
 		} else {
-			fmap["cacert"] = caPath
-			fmap["cert"] = certPath
-			fmap["key"] = privateKeyPath
+			fmap["cacert"] = e2e.CaPath
+			fmap["cert"] = e2e.CertPath
+			fmap["key"] = e2e.PrivateKeyPath
 		}
 	}
 	if cx.user != "" {
@@ -242,12 +349,11 @@ func (cx *ctlCtx) prefixArgs(eps []string) []string {
 
 	useEnv := cx.envMap != nil
 
-	cmdArgs := []string{ctlBinPath + "3"}
+	cmdArgs := []string{e2e.CtlBinPath + "3"}
 	for k, v := range fmap {
 		if useEnv {
 			ek := flags.FlagToEnv("ETCDCTL", k)
-			os.Setenv(ek, v)
-			cx.envMap[ek] = struct{}{}
+			cx.envMap[ek] = v
 		} else {
 			cmdArgs = append(cmdArgs, fmt.Sprintf("--%s=%s", k, v))
 		}
@@ -261,12 +367,22 @@ func (cx *ctlCtx) PrefixArgs() []string {
 	return cx.prefixArgs(cx.epc.EndpointsV3())
 }
 
+// PrefixArgsUtl returns prefix of the command that is either etcdctl or etcdutl
+// depending on cx configuration.
+// Please not thet 'utl' compatible commands does not consume --endpoints flag.
+func (cx *ctlCtx) PrefixArgsUtl() []string {
+	if cx.etcdutl {
+		return []string{e2e.UtlBinPath}
+	}
+	return []string{e2e.CtlBinPath}
+}
+
 func isGRPCTimedout(err error) bool {
 	return strings.Contains(err.Error(), "grpc: timed out trying to connect")
 }
 
 func (cx *ctlCtx) memberToRemove() (ep string, memberID string, clusterID string) {
-	n1 := cx.cfg.clusterSize
+	n1 := cx.cfg.ClusterSize
 	if n1 < 2 {
 		cx.t.Fatalf("%d-node is too small to test 'member remove'", n1)
 	}
