@@ -19,17 +19,19 @@ import (
 	"fmt"
 	"strings"
 
-	"go.etcd.io/etcd/clientv3"
+	clientv3 "go.etcd.io/etcd/client/v3"
+	"go.etcd.io/etcd/tests/v3/framework/e2e"
+	"go.etcd.io/etcd/tests/v3/integration"
 )
 
 type Etcdctl struct {
-	connType  clientConnType
+	connType  e2e.ClientConnType
 	isAutoTLS bool
 	endpoints []string
 	v2        bool
 }
 
-func NewEtcdctl(endpoints []string, connType clientConnType, isAutoTLS bool, v2 bool) *Etcdctl {
+func NewEtcdctl(endpoints []string, connType e2e.ClientConnType, isAutoTLS bool, v2 bool) *Etcdctl {
 	return &Etcdctl{
 		endpoints: endpoints,
 		connType:  connType,
@@ -40,7 +42,7 @@ func NewEtcdctl(endpoints []string, connType clientConnType, isAutoTLS bool, v2
 
 func (ctl *Etcdctl) Get(key string) (*clientv3.GetResponse, error) {
 	var resp clientv3.GetResponse
-	err := ctl.spawnJsonCmd(&resp, "", "get", key)
+	err := ctl.spawnJsonCmd(&resp, "get", key)
 	return &resp, err
 }
 
@@ -50,7 +52,7 @@ func (ctl *Etcdctl) Put(key, value string) error {
 	}
 	args := ctl.cmdArgs()
 	args = append(args, "put", key, value)
-	return spawnWithExpect(args, "OK")
+	return e2e.SpawnWithExpectWithEnv(args, ctl.env(), "OK")
 }
 
 func (ctl *Etcdctl) PutWithAuth(key, value, username, password string) error {
@@ -59,7 +61,7 @@ func (ctl *Etcdctl) PutWithAuth(key, value, username, password string) error {
 	}
 	args := ctl.cmdArgs()
 	args = append(args, "--user", fmt.Sprintf("%s:%s", username, password), "put", key, value)
-	return spawnWithExpect(args, "OK")
+	return e2e.SpawnWithExpectWithEnv(args, ctl.env(), "OK")
 }
 
 func (ctl *Etcdctl) Set(key, value string) error {
@@ -68,17 +70,25 @@ func (ctl *Etcdctl) Set(key, value string) error {
 	}
 	args := ctl.cmdArgs()
 	args = append(args, "set", key, value)
-	return spawnWithExpect(args, value)
+	lines, err := e2e.RunUtilCompletion(args, ctl.env())
+	if err != nil {
+		return err
+	}
+	response := strings.ReplaceAll(strings.Join(lines, "\n"), "\r\n", "")
+	if response != value {
+		return fmt.Errorf("Got unexpected response %q, expected %q", response, value)
+	}
+	return nil
 }
 
 func (ctl *Etcdctl) AuthEnable() error {
 	args := ctl.cmdArgs("auth", "enable")
-	return spawnWithExpect(args, "Authentication Enabled")
+	return e2e.SpawnWithExpectWithEnv(args, ctl.env(), "Authentication Enabled")
 }
 
 func (ctl *Etcdctl) UserGrantRole(user string, role string) (*clientv3.AuthUserGrantRoleResponse, error) {
 	var resp clientv3.AuthUserGrantRoleResponse
-	err := ctl.spawnJsonCmd(&resp, "", "user", "grant-role", user, role)
+	err := ctl.spawnJsonCmd(&resp, "user", "grant-role", user, role)
 	return &resp, err
 }
 
@@ -93,7 +103,7 @@ func (ctl *Etcdctl) UserAdd(name, password string) (*clientv3.AuthUserAddRespons
 	args = append(args, "--interactive=false")
 
 	var resp clientv3.AuthUserAddResponse
-	err := ctl.spawnJsonCmd(&resp, "", args...)
+	err := ctl.spawnJsonCmd(&resp, args...)
 	return &resp, err
 }
 
@@ -102,7 +112,7 @@ func (ctl *Etcdctl) AlarmList() (*clientv3.AlarmResponse, error) {
 		panic("Unsupported method for v2")
 	}
 	var resp clientv3.AlarmResponse
-	err := ctl.spawnJsonCmd(&resp, "{", "alarm", "list")
+	err := ctl.spawnJsonCmd(&resp, "alarm", "list")
 	return &resp, err
 }
 
@@ -111,7 +121,7 @@ func (ctl *Etcdctl) MemberList() (*clientv3.MemberListResponse, error) {
 		panic("Unsupported method for v2")
 	}
 	var resp clientv3.MemberListResponse
-	err := ctl.spawnJsonCmd(&resp, "", "member", "list")
+	err := ctl.spawnJsonCmd(&resp, "member", "list")
 	return &resp, err
 }
 
@@ -120,7 +130,7 @@ func (ctl *Etcdctl) MemberAdd(name string, peerURLs []string) (*clientv3.MemberA
 		panic("Unsupported method for v2")
 	}
 	var resp clientv3.MemberAddResponse
-	err := ctl.spawnJsonCmd(&resp, "", "member", "add", name, "--peer-urls", strings.Join(peerURLs, ","))
+	err := ctl.spawnJsonCmd(&resp, "member", "add", name, "--peer-urls", strings.Join(peerURLs, ","))
 	return &resp, err
 }
 
@@ -129,7 +139,7 @@ func (ctl *Etcdctl) MemberRemove(id uint64) (*clientv3.MemberRemoveResponse, err
 		panic("Unsupported method for v2")
 	}
 	var resp clientv3.MemberRemoveResponse
-	err := ctl.spawnJsonCmd(&resp, "", "member", "remove", fmt.Sprintf("%x", id))
+	err := ctl.spawnJsonCmd(&resp, "member", "remove", fmt.Sprintf("%x", id))
 	return &resp, err
 }
 
@@ -138,35 +148,16 @@ func (ctl *Etcdctl) Compact(rev int64) (*clientv3.CompactResponse, error) {
 		panic("Unsupported method for v2")
 	}
 	args := ctl.cmdArgs("compact", fmt.Sprint(rev))
-	return nil, spawnWithExpect(args, fmt.Sprintf("compacted revision %v", rev))
+	return nil, e2e.SpawnWithExpectWithEnv(args, ctl.env(), fmt.Sprintf("compacted revision %v", rev))
 }
 
-func (ctl *Etcdctl) Status() ([]*clientv3.StatusResponse, error) {
-	var epStatus []*struct {
-		Endpoint string
-		Status   *clientv3.StatusResponse
-	}
-	err := ctl.spawnJsonCmd(&epStatus, "", "endpoint", "status")
-	if err != nil {
-		return nil, err
-	}
-	resp := make([]*clientv3.StatusResponse, len(epStatus))
-	for i, e := range epStatus {
-		resp[i] = e.Status
-	}
-	return resp, err
-}
-
-func (ctl *Etcdctl) spawnJsonCmd(output interface{}, expectedOutput string, args ...string) error {
+func (ctl *Etcdctl) spawnJsonCmd(output interface{}, args ...string) error {
 	args = append(args, "-w", "json")
-	cmd, err := spawnCmd(append(ctl.cmdArgs(), args...))
+	cmd, err := e2e.SpawnCmd(append(ctl.cmdArgs(), args...), ctl.env())
 	if err != nil {
 		return err
 	}
-	if expectedOutput == "" {
-		expectedOutput = "header"
-	}
-	line, err := cmd.Expect(expectedOutput)
+	line, err := cmd.Expect("header")
 	if err != nil {
 		return err
 	}
@@ -174,13 +165,7 @@ func (ctl *Etcdctl) spawnJsonCmd(output interface{}, expectedOutput string, args
 }
 
 func (ctl *Etcdctl) cmdArgs(args ...string) []string {
-	var cmdArgs []string
-	binPath := ctlBinPath
-	if ctl.v2 {
-		cmdArgs = []string{binPath + "2"}
-	} else {
-		cmdArgs = []string{binPath + "3"}
-	}
+	cmdArgs := []string{e2e.CtlBinPath}
 	for k, v := range ctl.flags() {
 		cmdArgs = append(cmdArgs, fmt.Sprintf("--%s=%s", k, v))
 	}
@@ -191,23 +176,33 @@ func (ctl *Etcdctl) flags() map[string]string {
 	fmap := make(map[string]string)
 	if ctl.v2 {
 		fmap["no-sync"] = "true"
-		if ctl.connType == clientTLS {
-			fmap["ca-file"] = testTLSInfo.TrustedCAFile
-			fmap["cert-file"] = testTLSInfo.CertFile
-			fmap["key-file"] = testTLSInfo.KeyFile
+		if ctl.connType == e2e.ClientTLS {
+			fmap["ca-file"] = integration.TestTLSInfo.TrustedCAFile
+			fmap["cert-file"] = integration.TestTLSInfo.CertFile
+			fmap["key-file"] = integration.TestTLSInfo.KeyFile
 		}
 	} else {
-		if ctl.connType == clientTLS {
+		if ctl.connType == e2e.ClientTLS {
 			if ctl.isAutoTLS {
 				fmap["insecure-transport"] = "false"
 				fmap["insecure-skip-tls-verify"] = "true"
 			} else {
-				fmap["cacert"] = testTLSInfo.TrustedCAFile
-				fmap["cert"] = testTLSInfo.CertFile
-				fmap["key"] = testTLSInfo.KeyFile
+				fmap["cacert"] = integration.TestTLSInfo.TrustedCAFile
+				fmap["cert"] = integration.TestTLSInfo.CertFile
+				fmap["key"] = integration.TestTLSInfo.KeyFile
 			}
 		}
 	}
 	fmap["endpoints"] = strings.Join(ctl.endpoints, ",")
 	return fmap
 }
+
+func (ctl *Etcdctl) env() map[string]string {
+	env := make(map[string]string)
+	if ctl.v2 {
+		env["ETCDCTL_API"] = "2"
+	} else {
+		env["ETCDCTL_API"] = "3"
+	}
+	return env
+}
